package se.chalmers.cs.gf.tramdemo;

import java.awt.Point;
import java.util.Collection;
import java.util.Collections;
import java.util.EventObject;

public class MapClickEvent extends EventObject {

        private Collection<Point> points;

        private Collection<String> regions;

        private long time;

        public MapClickEvent(Object source, Collection<Point> points, Collection<String> regions, long time) {
                super(source);
                this.points = Collections.unmodifiableCollection(points);
                this.regions = Collections.unmodifiableCollection(regions);
                this.time = time;
        }


        public MapClickEvent(Object source, Collection<Point> points, long time) {
                this(source, points, Collections.<String>emptyList(), time);
        }


        public MapClickEvent(Object source, Point point, Collection<String> regions, long time) {
                this(source, Collections.<Point>singleton(point), regions, time);
        }

        public MapClickEvent(Object source, Point point, long time) {
                this(source, point, Collections.<String>emptyList(), time);
        }

        /**
         *  Gets the map coordinates of the clicks.
         */
        public Collection<Point> getPoints() {
                return points;
        }

        public Collection<String> getRegions() {
                return regions;
        }

        /**
         *  Get the time (in milliseconds since the Unix epoch)
         *  at which the click happend.
         */
        public long getTime() {
                return time;
        }

        public String toString() {

                StringBuilder sb = new StringBuilder("MapClick({");
                for (Point p : points) {
                        sb.append("(").append(p.x).append(",").append(p.y).append(")");
                }
                sb.append("},");
                sb.append(regions);
                sb.append(")");

                return sb.toString();
        }

}
