package se.chalmers.cs.gf.tramdemo;

import se.chalmers.cs.gf.util.Pair;
import static se.chalmers.cs.gf.util.Pair.pair;
import se.chalmers.cs.gf.abssyn.Tree;
import se.chalmers.cs.gf.oaa.TreeToIclTerm;
import se.chalmers.cs.gf.oaa.IclTermToTree;

import se.chalmers.cs.gf.dialogutil.*;
import se.chalmers.cs.gf.dialogutil.sr.Recognizer;
import se.chalmers.cs.gf.dialogutil.tts.*;

import se.chalmers.cs.gf.graph.Path;

import se.chalmers.cs.gf.tramdemo.query.Query;
import se.chalmers.cs.gf.tramdemo.query.QInput;
import se.chalmers.cs.gf.tramdemo.query.GoFromTo;
import se.chalmers.cs.gf.tramdemo.query.PNamed;
import se.chalmers.cs.gf.tramdemo.query.TreeToTyped;

import se.chalmers.cs.gf.tramdemo.answer.Route;
import se.chalmers.cs.gf.tramdemo.answer.TypedToTree;

import com.sri.oaa2.icl.*;

import java.io.IOException;
import java.util.*;

public class OAATramDemo implements Runnable {

        /** Time allowed for clicks after end of text input (in milliseconds). */
        private static final int CLICK_TIMEOUT = 1000;

        private OAAClient oaa;

        private Recognizer sr;

        private TextListenerList outputs = new TextListenerList();

        public OAATramDemo(boolean useSR, OAAClient oaa) {
                this.oaa = oaa;
                if (useSR) {
                        this.sr = new Recognizer(oaa);
                } else {
                        this.sr = null;
                }

                outputs.add(new FreeTTSAgentSpeechOutput(oaa));
                outputs.add(new NuanceWrapperSpeechOutput(oaa));
        }

        public void run() {
                while (true) {

                        // get input

                        String textInput = getTextInput();

                        System.err.println("TramDemo: Text input: " + textInput);

                        delay(CLICK_TIMEOUT);

                        List<Click> clickInput = getClickInput();

                        System.err.println("TramDemo: Clicks: " + clickInput);

                        String multiModalInput 
                                = combineTextAndClicks(textInput, clickInput);

                        System.err.println("TramDemo: Combined input: " + multiModalInput);

                        // parse and interpret input

                        Collection<Pair<String,Query>> inputs = parseQuery(multiModalInput);

                        System.err.println("TramDemo: Parse results: " + inputs);

                        List<Pair<String,Set<ShortestPathQuery>>> queries = interpretQueries(inputs);

                        System.err.println("TramDemo: Interpreted inputs: " + queries);

                        Pair<String,Set<ShortestPathQuery>> input = disambiguate(queries);
                        if (input == null)
                                continue;

                        System.err.println("TramDemo: Disambiguated input: " + input);

                        String inputLang = input.fst;
                        Set<ShortestPathQuery> qs = input.snd;

                        // answer query

                        Route route = findRoute(qs);

                        System.err.println("TramDemo: Route: " + route);

                        // draw output

                        String mapLang = "GbgRouteMap"; // FIXME: figure out dynamically

                        String drawInstrs = linearizeRoute(mapLang, route);

                        System.err.println("TramDemo: Drawing: " + drawInstrs);

                        draw(drawInstrs);

                        // speak output

                        String outputLang = getOutputLang(inputLang);

                        System.err.println("TramDemo: Output language: " + outputLang);

                        String outputText = linearizeRoute(outputLang, route);

                        System.err.println("TramDemo: Output text: " + outputText);

                        say(outputLang, outputText);

                }
        }
        
        private String getTextInput() {
                while (true) {
                        if (sr == null) {
                                String r = javax.swing.JOptionPane.showInputDialog("Query");
                                if (r != null)
                                        return r;
                        } else {
                                Recognizer.Result r = sr.recognize(".MAIN");
                                if (r != null)
                                        return r.getText();
                        }
                }
        }

        private List<Click> getClickInput() {
                IclTerm goal = IclUtil.icl("clicks(Clicks)");
                IclList clicks = (IclList)solveAndGetVar(goal, "Clicks");

                List<Click> l = new LinkedList<Click>();
                //click(['Chalmers','Vasaplatsen'])
                for (IclTerm click : IclUtil.toList(clicks)) {
                        Click c = new Click();
                        l.add(c);
                        for (IclTerm region : IclUtil.toList((IclList)click.getTerm(0)))
                                c.add(IclUtil.fromStr(region));
                }

                return l;
        }

        private String combineTextAndClicks(String textInput, List<Click> clickInput) {
                StringBuilder r = new StringBuilder(textInput);
                r.append(";");
                for (Click c : clickInput) {
                        r.append(" [");
                        Iterator<String> it = c.iterator();
                        while (it.hasNext()) {
                                // FIXME: do real escaping
                                String esc = '"' + it.next() + '"';
                                r.append(esc);
                                if (it.hasNext())
                                        r.append(",");
                        }
                        r.append("]");
                }
                return r.toString();
        }

        private Collection<Pair<String,Query>> parseQuery(String multiModalInput) {
                IclTerm goal = new IclStruct("parse", 
                                             new IclStr("query"),
                                             new IclVar("Lang"), 
                                             new IclStr(multiModalInput), 
                                             new IclVar("Tree"));
                List<IclTerm> answers = oaa.solve(goal);
                List<Pair<String,Query>> l = new LinkedList<Pair<String,Query>>();
                for (IclTerm answer : answers) {
                        Map<String,IclTerm> vars = IclUtil.matchAndGetVars(goal, answer);
                        String lang = IclUtil.fromStr(vars.get("Lang"));
                        IclTerm treeTerm = vars.get("Tree");
                        Tree tree = IclTermToTree.iclTermToTree(treeTerm);
                        Query query = TreeToTyped.getInstance().toQuery(tree);
                        l.add(pair(lang, query));
                }

                return l;
        }

        private List<Pair<String,Set<ShortestPathQuery>>> interpretQueries(
                Collection<Pair<String,Query>> qs) {

                List<Pair<String,Set<ShortestPathQuery>>> l = 
                        new LinkedList<Pair<String,Set<ShortestPathQuery>>>();
                for (Pair<String,Query> p : qs)
                        l.add(pair(p.fst,interpretQuery(p.snd)));

                return l;
        }

        private Set<ShortestPathQuery> interpretQuery(Query q) {
                Tree tree = se.chalmers.cs.gf.tramdemo.query.TypedToTree.getInstance().toTree(q);
                IclTerm term = TreeToIclTerm.treeToIclTerm(tree);

                IclTerm goal = new IclStruct("interpret_query",
                                             term,
                                             new IclStruct("shortest_path_query",
                                                           new IclVar("From"),
                                                           new IclVar("To")));
                List<IclTerm> rs = oaa.solve(goal);

                Set<ShortestPathQuery> l = new HashSet<ShortestPathQuery>();
                for (IclTerm r : rs) {
                        Map<String,IclTerm> m = IclUtil.matchAndGetVars(goal, r);
                        String from = IclUtil.fromStr(m.get("From"));
                        String to = IclUtil.fromStr(m.get("To"));
                        l.add(new ShortestPathQuery(from, to));
                }

                return l;
        }

        private <A> Pair<String,A> disambiguate(
                Collection<Pair<String,A>> inputs) {

                if (inputs.size() == 0) {
                        System.err.println("ERROR: no parse");
                        return null;
                } else if (inputs.size() == 1) {
                        return inputs.iterator().next();
                } else {
                        System.err.println("ERROR: ambiguous:");
                        for (Pair<String,A> q : inputs)
                                System.err.println(q.fst + ": " + q.snd);
                        return null;
                }
        }

        private Route findRoute(Collection<ShortestPathQuery> qs) {
                Path shortestPath = null;
                double shortestLength = Double.POSITIVE_INFINITY;

                for (ShortestPathQuery query : qs) {

                        IclTerm goal = new IclStruct("shortest_path",
                                                     new IclStr(query.getFrom()),
                                                     new IclStr(query.getTo()),
                                                     new IclVar("Path"));
                        IclList answer = (IclList)solveAndGetVar(goal, "Path");
                        
                        Path path = new Path();
                        for (IclTerm legTerm : IclUtil.toList(answer)) {
                                // leg('6','Valand','Chalmers', 3.0)
                                String label = IclUtil.fromStr(legTerm.getTerm(0));
                                String from =  IclUtil.fromStr(legTerm.getTerm(1));
                                String to =    IclUtil.fromStr(legTerm.getTerm(2));
                                double dist =  ((IclFloat)legTerm.getTerm(3)).toDouble();
                                path.addLeg(from, to, dist, label);
                        }

                        double l = path.getLength();
                        if (l < shortestLength) {
                                shortestPath = path;
                                shortestLength = l;
                        }
                }

                return PathToRoute.pathToRoute(shortestPath);
        }

        private String getOutputLang(String inputLang) {
                String lang = inputLang.substring(inputLang.length()-3);
                return "GbgRoute" + lang; // FIXME: figure out city name dynamically
        }

        private String linearizeRoute(String lang, Route route) {
                Tree tree = TypedToTree.getInstance().toTree(route);
                IclTerm term = TreeToIclTerm.treeToIclTerm(tree);
                IclTerm goal = new IclStruct("linearize", 
                                             new IclStr("answer"),
                                             new IclStr(lang), 
                                             term, 
                                             new IclVar("Text"));
                IclStr answer = (IclStr)solveAndGetVar(goal, "Text");
                return answer.toUnquotedString();
        }

        private void say(String lang, String text) {
                outputs.fireTextEvent(text);
        }

        private void draw(String drawInstrs) {
                oaa.solve(IclUtil.icl("draw(clear)"));
                IclTerm goal = new IclStruct("draw" , new IclStr(drawInstrs));
                oaa.solve(goal);
        }


        /**
         *  Solve a goal and get the value of the given variable in the
         *  single result. Fails if there are no or multiple results.
         */
        private IclTerm solveAndGetVar(IclTerm goal, String var) {
                List<IclTerm> answers = oaa.solve(goal);
                if (answers.size() < 1)
                        throw new RuntimeException("Could not solve " + goal);
                if (answers.size() > 1)
                        throw new RuntimeException("Multiple results for " + goal);
                IclTerm answer = answers.get(0);
                return IclUtil.matchAndGetVar(goal, answer, var);
        }

        private void delay(long ms) {
                try {
                        Thread.sleep(ms);
                } catch (InterruptedException ex) { }
        }

        private static class Click extends LinkedList<String> { }

        public static void main(String[] args) {
                try {
                        boolean useSR = args.length > 0 && args[0].equals("-sr");
                        OAAClient oaa = new OAAClient("TramDemo", args);
                        new OAATramDemo(useSR, oaa).run();
                } catch (IOException ex) {
                        System.err.println(ex);
                }
        }

}
