package se.chalmers.cs.gf.tramdemo;

import se.chalmers.cs.gf.abssyn.Fun;

import se.chalmers.cs.gf.tramdemo.query.*;
import se.chalmers.cs.gf.tramdemo.answer.*;
import se.chalmers.cs.gf.graph.*;

import java.util.*;

public class PathToRoute {

        public static Route pathToRoute(Path p) {
                List<Leg> routeLegs = new LinkedList<Leg>();

                if (p.getLegs().isEmpty())
                        return new LegsRoute(new OneLeg(new LineLeg(new NamedLine("nothing"), 
                                                      new TwoStops(new se.chalmers.cs.gf.tramdemo.answer.Chalmers(),
                                                                   new se.chalmers.cs.gf.tramdemo.answer.Chalmers())))); // FIXME: hack

                for (List<Path.Leg> ls : groupLegs(p.getLegs())) {
                        assert !ls.isEmpty();

                        Path.Leg firstLeg = ls.get(0);
                        String label = firstLeg.getLabel();

                        List<String> stopNames = new LinkedList<String>();
                        stopNames.add(firstLeg.getFrom());
                        for (Path.Leg l : ls)
                                stopNames.add(l.getTo());

                        routeLegs.add(buildLineLeg(label, stopNames));
                }
  
                return buildRoute(routeLegs);
        }


        /**
         *  Group legs by line number.
         */
        private static List<List<Path.Leg>> groupLegs(List<Path.Leg> ls) {
                List<List<Path.Leg>> r = new LinkedList<List<Path.Leg>>();
                
                String currLabel = null;
                List<Path.Leg> currLegs = null;

                for (Path.Leg l : ls) {
                        if (currLegs == null || !l.getLabel().equals(currLabel)) {
                                currLegs = new LinkedList<Path.Leg>();
                                r.add(currLegs);
                                currLabel = l.getLabel();
                        }
                        currLegs.add(l);
                }


                return r;
        }

        private static Leg buildLineLeg(String label, List<String> stopNames) {
                Iterator<String> it = ReverseIterator.create(stopNames);
                // FIXME: what if there are not at least two stops?
                se.chalmers.cs.gf.tramdemo.answer.Stop last = makeStop(it.next());
                se.chalmers.cs.gf.tramdemo.answer.Stop secondLast = makeStop(it.next());
                Stops stops = new TwoStops(secondLast, last);
                while (it.hasNext()) {
                        se.chalmers.cs.gf.tramdemo.answer.Stop stop = makeStop(it.next());
                        stops = new ConsStop(stop, stops);
                }
                Line line = new NamedLine(label);
                return new LineLeg(line, stops);
        }

        private static Route buildRoute(List<Leg> routeLegs) {
                Iterator<Leg> it = ReverseIterator.create(routeLegs);
                Legs ls = new OneLeg(it.next()); // FIXME: what if there are no legs?
                while (it.hasNext())
                        ls = new Then(it.next(), ls);
                return new LegsRoute(ls);
        }

        private static se.chalmers.cs.gf.tramdemo.answer.Stop makeStop(String stopName) {
                return se.chalmers.cs.gf.tramdemo.answer.TreeToTyped.getInstance().toStop(new Fun(stopName));
        }

}
